package com.clusterapp.servlet;

import javax.servlet.*;
import javax.servlet.http.*;
import java.io.*;
import java.util.*;
import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;

import com.clusterapp.listeners.ClusterAppSessionListener;
import com.clusterapp.session.SessionAttributeFactory;

public class ClusterAppServlet extends HttpServlet {

	private static Log sLog = LogFactory.getLog(ClusterAppServlet.class);
	private static final String CONTENT_TYPE = "text/plain";

	// Initialize global variables
	public void init() throws ServletException {
		sLog.debug(getInitParameter("clusterNode") + "- init() method called.");
	}

	// Clean up resources
	public void destroy() {
		sLog.debug(getInitParameter("clusterNode") + "- destroy() method called.");
	}

	public void service(HttpServletRequest request, HttpServletResponse response)
			throws ServletException {
		sLog.debug(getInitParameter("clusterNode") + "- service() method called.");

//		if (req.getSession().isNew() && (req.getParameter("sessionId") == null)) {
//			sLog.debug("Create a new HTTP session");
//			String response = generateNewSession(req, res, true);
//			sLog.debug("response : " + response);
//			try {
//			res.getOutputStream().write(response.getBytes());
//			} catch (IOException ioe) {
//			    sLog.error(ioe);
//			    throw new ServletException(ioe);
//			}
//		} else {
//			sLog.debug("Evaluating existing HTTP session");
//			String response = evaluateSession(req, res);
//			sLog.debug("response : " + response);
//		}

		String sessionCommand = request.getParameter("sessionCommand");
		sLog.debug("**** sessionCommand: " + sessionCommand);
		String strResponse = "";
		// create a new session
		if ("create".equals(sessionCommand)) {
			sLog.debug("Create a new HTTP session");
			strResponse = generateNewSession(request, response);
		} else if ("expire".equals(sessionCommand)) { // invalidate the session
			sLog.debug("Expire the HTTP session");
			sLog.debug("SessionCount = " + ClusterAppSessionListener.getSessionCount());
			if (ClusterAppSessionListener.getSessionCount() == 0) {
				generateErrorReport(request, response, "No sessions to expire.");
				return;
			}
			request.getSession().invalidate();
			strResponse = generateContinuePage(request, response);
		} else {
			sLog.debug("Evaluating existing HTTP session");
			strResponse = evaluateSession(request, response);
			sLog.debug("response : " + response);
		}
		sLog.debug("strResponse : " + strResponse);
		try {
			response.getOutputStream().write(strResponse.getBytes());
		} catch (IOException ioe) {
			sLog.error(ioe);
			throw new ServletException(ioe);
		}
	}

	public synchronized String generateNewSession(HttpServletRequest request,
			HttpServletResponse response) throws ServletException {
//			HttpServletResponse res, boolean integritycheck) throws ServletException {
		sLog.debug(getInitParameter("clusterNode") + "- generateNewSession() method called.");
//		if (integritycheck) {
			String id = request.getParameter("sessionId");
			sLog.debug("sessionId=" + id);
			if (id != null) {
				sLog.debug("Started a new session with an existing client session. Client session="
				+ id + " server session=" + request.getSession().getId());
				generateErrorReport(request, response,
						"Started a new session with an existing client session. Client session="
								+ id + " server session=" + request.getSession().getId());
				return "";
			}
//		} else {
//			req.getSession().invalidate();
//		}
		response.setContentType(CONTENT_TYPE);
		request.getSession().setAttribute("sessionCounter", "1");
		return generateContinuePage(request, response);
	}

	public synchronized String evaluateSession(HttpServletRequest request,
			HttpServletResponse response) throws ServletException {
		sLog.debug(getInitParameter("clusterNode") + "- evaluateSession() method called.");
		response.setContentType(CONTENT_TYPE);
		String sessionCommand = request.getParameter("sessionCommand");
		// check if we want to create a new session
//		sLog.debug("**** sessionCommand: " + sessionCommand);
//		// create a new session
//		if ("create".equals(sessionCommand)) {
//			return generateNewSession(req, res, false);
//		}

		String id = request.getParameter("sessionId");
		sLog.debug("id" + id);
		HttpSession session = request.getSession();
		if (!session.getId().equals(id)) {
			generateErrorReport(request, response, "Invalid session id server=" + session.getId() + " client=" + id);
			return "";
		}

		String counter = request.getParameter("sessionCounter");
		String c = (String) session.getAttribute("sessionCounter");
		if (c == null || (!c.equals(counter))) {
			generateErrorReport(request, response, "Invalid counter id server=" + c + " client=" + counter);
			return "";
		} //end if
		c = String.valueOf(Integer.parseInt(c) + 1);
		session.setAttribute("sessionCounter", c);

		//loop through client parameters
		java.util.Enumeration e = request.getParameterNames();
		while (e.hasMoreElements()) {
			String name = (String) e.nextElement();
			if (name.startsWith("param_")) {
				String value = request.getParameter(name);
				Object svalue = session.getAttribute(name);
				if (svalue == null || value == null) {
					generateErrorReport(request, response, "Session data mismatch[1] client name=" + name
									+ " value=" + value + " server value=" + svalue);
					return "";
				}
				if (!value.equals(svalue.toString())) {
					generateErrorReport(request, response, "Session data mismatch[2] client name=" + name
									+ " value=" + value + " server value=" + svalue);
					return "";
				}
			}//end if
		}//while
		//loop through server attributes to match
		e = session.getAttributeNames();
		while (e.hasMoreElements()) {
			String name = (String) e.nextElement();
			if (name.startsWith("param_")) {
				String value = request.getParameter(name);
				Object svalue = session.getAttribute(name);
				if (svalue == null || value == null) {
					generateErrorReport(request, response, "Session data mismatch[3] client name=" + name
									+ " value=" + value + " server value=" + svalue);
					return "";
				}
				if (!value.equals(svalue.toString())) {
					generateErrorReport(request, response, "Session data mismatch[4] client name=" + name
									+ " value=" + value + " server value=" + svalue);
					return "";
				}
			}//end if
		}//while

//		String command = req.getParameter("sessionCommand");
//		sLog.debug("**** command: " + command);

		// check for null session command
		if (sessionCommand == null) {
			generateErrorReport(request, response, "Session Command is null, so do nothing.");
			return "";
		}
		String attrName = request.getParameter("sessionAttrName");
		String attrValue = request.getParameter("sessionAttrValue");
		if (attrName == null || attrValue == null || session.getAttribute(attrName) == null) {
			generateErrorReport(request, response, "Invalid attribute name or value.");
			return "";
		}
//		} else if ("remove".equals(sessionCommand)) {
//			if (session.getAttribute(req.getParameter("sessionAttrName")) == null) {
//				generateErrorReport(req, res,
//						"Unable to remove attribute, nonexistent client name="
//								+ req.getParameter("sessionAttrName"));
//				return "";
//			} else {
//				sLog.debug("[" + id + "] Attribute removed name="
//						+ req.getParameter("sessionAttrName")
//						+ " Value=" + session.getAttribute(attrName));
//				session.removeAttribute(attrName);
//			}//end if
//		}//end if

		// add a session attribute
		if ("add".equals(sessionCommand)) {
			session.setAttribute(attrName, SessionAttributeFactory.getSessionAttribute(attrValue));
			sLog.debug("[" + id + "] Added attribute: Name= " + attrName +
					" Value=" + session.getAttribute(attrName));
		}
		// remove a session attribute
		if ("remove".equals(sessionCommand)) {
//			if (session.getAttribute(attrName) == null) {
//				generateErrorReport(req, res, "Unable to remove attribute, nonexistent client name=" + attrName);
//				return "";
//			} else {
				session.removeAttribute(attrName);
				sLog.debug("[" + id + "] Removed attribute: Name = " + attrName +
						" Value=" + session.getAttribute(attrName));
//			}//end if
		}
//		// invalidate the session
//		if ("expire".equals(sessionCommand)) {
//			req.getSession().invalidate();
//		}
		return generateContinuePage(request, response);
	}

	public void generateErrorReport(HttpServletRequest request,
			HttpServletResponse response, String message) throws ServletException {
		sLog.debug(getInitParameter("clusterNode") + "- generateErrorReport() method called.");
		// When an error happens print out headers and more
		StringBuffer strRequest = new StringBuffer("Error request on session:").append(message);
		strRequest.append("\nHEADERS:");
		java.util.Enumeration e = request.getHeaderNames();
		while (e.hasMoreElements()) {
			String name = (String) e.nextElement();
			strRequest.append("\n\t").append(name).append("=").append(request.getHeader(name));
		}
		strRequest.append("\nPARAMETERS:");
		e = request.getParameterNames();
		while (e.hasMoreElements()) {
			String name = (String) e.nextElement();
			strRequest.append("\n\t").append(name).append("=").append(request.getParameter(name));
		}
		strRequest.append("\n");
		sLog.debug(request);
		try {
			String msg = "[" + java.net.InetAddress.getLocalHost() + "]" + message;
			response.sendError(501, msg);
		} catch (Exception x) {
			x.printStackTrace();
		}
	}

	public String generateContinuePage(HttpServletRequest request,
			HttpServletResponse response) throws ServletException {
		/**
		 * generates a java.util.Properties format response. with the following parameters
		 * sessionId
		 * param_xx - where xx is a counter on the session attribute name
		 * value_xx - where xx is a counter on the session attribute value
		 * counter  - the number of requests
		 */
		sLog.debug(getInitParameter("clusterNode") + "- generateContinuePage() method called.");
		java.util.Properties p = new java.util.Properties();
		java.util.Enumeration e = request.getSession().getAttributeNames();
		java.io.ByteArrayOutputStream out = new java.io.ByteArrayOutputStream();
		p.setProperty("sessionId", request.getSession().getId());
		while (e.hasMoreElements()) {
			String name = (String) e.nextElement();
			p.setProperty(name, request.getSession().getAttribute(name).toString());
		}
		try {
			p.store(out, "");
			return new String(out.toByteArray());
		} catch (Exception x) {
			throw new ServletException(x);
		}
	}
}
