package com.clusterapp.test;

import org.apache.commons.httpclient.HttpClient;
import org.apache.commons.httpclient.NameValuePair;
import org.apache.commons.httpclient.Header;
import org.apache.commons.httpclient.methods.PostMethod;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;

import com.clusterapp.session.SessionAttribute;
import com.clusterapp.session.SessionAttributeFactory;
import com.clusterapp.session.BigSessionAttribute;
import com.clusterapp.session.LiteSessionAttribute;

public class SessionReplicationClient extends Thread {

	private static Log sLog = LogFactory.getLog(SessionReplicationClient.class);

	public static String[] testdata = {
			"TestAttribute1", "TestAttribute2",
			"TestAttribute3", "TestAttribute4", "TestAttribute5",
			"TestAttribute6", "TestAttribute7", "TestAttribute8",
			"TestAttribute9","TestAttribute10"};

	private String host;
	private int port;
	private String webPage = "clusterapp/ReplicationServlet";
	private int reps;
	private long delay;
	private static String attrType;

	private int counter = 0;
	private String sessionCommand = "add";
	private java.util.Properties p = new java.util.Properties();
	private java.util.Properties previousp = new java.util.Properties();

	private static long responsetime = 0;
	private static long numOfRequests = 0;
	private static long lastcheck = System.currentTimeMillis();
	private static long maxResponseTime = -1;
	private static long minResponseTime = Long.MAX_VALUE;
	private static int maxThreadCount = 0;
	private static int deadThreadCount = 0;

	private static int threadCounter = 0;
	private static int[] errorCounter = new int[1000];

	static {
		java.util.Arrays.fill(errorCounter, 0);
	}

	// CONSTRUCTOR
	public SessionReplicationClient(String name, String host, int port,
			int reps, long delay, String attrType) {
		super(name);
		this.host = host;
		this.reps = reps;
		this.port = port;
		this.delay = delay;
		this.attrType = attrType;
	}

	public void run() {
		try {
			inc();
			runTest(host, port, reps, delay, attrType);
			SessionReplicationClient t = new SessionReplicationClient(this.getName(), host,
					port, reps, delay, attrType);
			t.setDaemon(false);
			t.start();
			sLog.debug("Thread " + this.getName() + " finished, started another one");
		} finally {
			dec();
		}
		printStats();
	}

	public void runTest(String host, int port, int reps, long delay, String attrType) {
		java.util.Random r = new java.util.Random(System.currentTimeMillis());
		HttpClient httpClient = new HttpClient();

		int allOK = 1;
		int counter = 0;
		for (int i = 0; (i < reps) && (allOK >= 0); i++) {
			boolean doAdd = r.nextBoolean();
			if (p.size() > 20) {
				doAdd = false;
			}
			int index = (r.nextInt() % 8);
			if (index < 0) {
				index = (index * ( -1));
			}
			if (allOK == 0) {
				p.setProperty("sessionCommand", "reset");
				sLog.debug("[" + Thread.currentThread().getName() +
						"] Sending reset command:");
			} else if (doAdd) {
//				p.setProperty("sessionCommand", "add");
//				p.setProperty("sessionCommandName", "param_" + i);
//				SessionAttribute attr = new SessionAttribute();
//				attr.setValues(new String[] {testdata[index]});
				// p.setProperty("sessionCommandValue", attr.getStringRepresentation());
//				p.setProperty("sessionCommandValue", attr.toString());

					sLog.debug("runTest(): sessionCommand=add");
					p.setProperty("sessionCommand", "add");
					//p.setProperty("sessionAttrName", "param_" + i);
					// Create a heavy or light-weight attribute based on the
					// command line argument
					SessionAttribute attr = null;
					if (attrType.equals("lite")) {
						// add 10 attributes
						for (int j=0; j<10; j++) {
							attr = new LiteSessionAttribute();
							attr.setValues(new String[] {testdata[j]});
//							p.setProperty("sessionAttrName"+j, "param_" + (i+j));
							p.setProperty("sessionAttrValue"+j, attr.toString());
						}
					} else {
//						attr = new BigSessionAttribute();
//						p.setProperty("sessionAttrName", "param_" + i);
//						p.setProperty("sessionAttrValue", attr.toString());

					   long start = System.currentTimeMillis();

//						for (int j=0; j<5; j++) {
							attr = new BigSessionAttribute();
							//attr.setValues(new String[] {testdata[j]});
							attr.setValues();
//							p.setProperty("sessionAttrName"+j, "param_" + (i+j));
//							p.setProperty("sessionAttrValue"+j, attr.toString());
							p.setProperty("sessionAttrValue", attr.toString());
//						}

						long stop = System.currentTimeMillis();
//						System.out.println("Time taken to create instance of BigSessionAttribute="+(stop-start));
						sLog.debug("Time taken to create instance of BigSessionAttribute="+(stop-start));

					}

			} else {
				p.setProperty("sessionCommand", "remove");
				int cc = 1;
				while ((p.getProperty("param_" + cc) == null)) {
					cc++;
					if (cc > 100000) {
						break;
					}
				}
				if (p.getProperty("param_" + cc) == null) {
					p.setProperty("sessionCommand", "nothing");
				} else {
					//sLog.debug("Found property to remove="+p.getProperty("param_"+(cc)));
					p.setProperty("sessionCommandName", "param_" + (cc));
				} //end if
			}
			allOK = this.doRequest(httpClient);
			counter++;
			try {
				//sLog.debug("Thread ["+Thread.currentThread().getName()+"] has received sessionId="+p.getProperty("sessionId"));
				Thread.sleep(delay);
			} catch (Exception ignore) {}
		} //for
		try {
			sLog.debug("Thread [" + Thread.currentThread().getName() +
					"] completed at index:" + counter);
			httpClient.endSession();
		} catch ( Exception x ) {
			x.printStackTrace();
		}

	}

	/**
	 *
	 * @param con
	 * @return 1 = all is ok
	 * @return 0 = try previous request
	 * @return -1 = failed
	 */
	public int doRequest(HttpClient httpClient) {
		try {
			NameValuePair form_data[] = new NameValuePair[p.size()];
			java.util.Enumeration e = p.propertyNames();
			int index = 0;
			while (e.hasMoreElements()) {
				String name = (String) e.nextElement();
				String value = p.getProperty(name);
				form_data[index++] = new NameValuePair(name, value);
			} //while
			//sLog.debug("Thread ["+Thread.currentThread().getName()+"] is sending sessionId="+p.getProperty("sessionId"));
			String strUrl = "http://" + host + ":" + port + "/" + webPage;
			sLog.debug("strUrl : " + strUrl);
			PostMethod method = new PostMethod(strUrl);
			method.setHttp11(false);
			method.setRequestBody(form_data);

			// note the convenience constructor for applets
			int statusCode = -1;
			int retries = 0;
			while (statusCode == -1) {
				try {
					sLog.debug("["+Thread.currentThread().getName()+"] Executing...");
					sLog.debug("["+Thread.currentThread().getName()+"] Request for SessionId="+p.getProperty("sessionId"));
					long start = System.currentTimeMillis();
					statusCode = httpClient.executeMethod(method);
					long stop = System.currentTimeMillis();
					addRspTime(stop - start, statusCode);
				} catch (Exception x) {
					try {
						Thread.currentThread().sleep(delay);
					} catch (Exception ignore) {}
					try {
						method.recycle();
						Thread.currentThread().sleep(10 * 1000);
					} catch (Exception ignore) {}
					if (retries > 10) {
						sLog.debug("[" + Thread.currentThread().getName() +
								"] Unable to execute command after 10 retries. Quitting:" +
						x.getMessage());
						x.printStackTrace();
						method.releaseConnection();
						return -1;
					}
				}
				retries++;
			}
			sLog.debug("statusCode : " + statusCode);
			return processStatusCode(statusCode, method);
		} catch (Exception ioe) {
			sLog.debug("[" + Thread.currentThread().getName() +
					"] Unable to receive data, lets try again:" +
			ioe.toString());
			return 0;
		}
	}

	private int processStatusCode(int statusCode, PostMethod method) throws Exception {
		if (statusCode == 302) {
			sLog.debug("[" + Thread.currentThread().getName() + "] Received 302 Error, \n"
					+ printRespHeader(method.getResponseHeaders()) + " will resend same command again.");
			String s = method.getResponseBodyAsString();
			sLog.debug("ResponseBody :" + s);
			incError(302);
			method.releaseConnection();
			return 1;
		} else if (statusCode == 400) {
			sLog.debug("[" + Thread.currentThread().getName() + "] Received Error: " + method.getStatusLine() + " "
					+ method.getStatusText() + " " + method.getResponseBodyAsString()
					+ "\nWill try again...after 1 sec sleep");
			String s = method.getResponseBodyAsString();
			sLog.debug("ResponseBody :" + s);
			try {
				Thread.currentThread().sleep(1000);
			} catch (Exception x) {
			}
			incError(statusCode);
			method.releaseConnection();
			return 1;
		} else if (statusCode >= 300) {
			sLog.debug("[" + Thread.currentThread().getName() + "] Received Error: " + method.getStatusLine() + " "
					+ method.getStatusText() + " " + method.getResponseBodyAsString());
			String s = method.getResponseBodyAsString();
			sLog.debug("ResponseBody :" + s);
			incError(statusCode);
			method.releaseConnection();
			return -1;
		} else if (statusCode == 200) {
			String s = method.getResponseBodyAsString();
			sLog.debug("ResponseBody :" + s);
			java.io.StringBufferInputStream in = new java.io.StringBufferInputStream(s);
			java.util.Properties tmp = new java.util.Properties();
			tmp.load(in);
			if (tmp.getProperty("sessionId") != null) {
				p = tmp;
				incError(statusCode);
				method.releaseConnection();
				return 1;
			} else {
				incError(601);
				//sLog.debug("Received invalid (wrong page) response from
				// server="+statusCode+" trying again...");
				//if ( p.getProperty("sessionId")==null ) {
				//	sLog.debug("No session id in the properties");
				//	System.exit(0);
				//}
				method.releaseConnection();
				return 1;
			}
			//sLog.debug("["+Thread.currentThread().getName()+"] completed
			// request #"+p.getProperty("counter")+" Number of
			// values="+p.size()+"\nResponse="+s);
		} else {
			sLog.debug("[" + Thread.currentThread().getName() + "] Received unknown response=" + statusCode + ", \n"
					+ printRespHeader(method.getResponseHeaders()) + " will resend same command again.");
			method.releaseConnection();
			return -1;
		}
	}

	private synchronized static void addRspTime(long time, int statusCode) {
		if (statusCode == 200) {
			maxResponseTime = Math.max(time, maxResponseTime);
			minResponseTime = Math.min(time, minResponseTime);
		}
		responsetime += time;
		numOfRequests++;
		if ( (numOfRequests % 100) == 0) {
			printStats();
			lastcheck = System.currentTimeMillis();
		}
	}

	private static String getErrorCodes() {
		StringBuffer buf = new StringBuffer();
		for (int i = 0; i < errorCounter.length; i++) {
			if (errorCounter[i] > 0) {
				buf.append("\n\tNr of ").append(i).append(" errors=").append(
				errorCounter[i]);
			}
		}
		return buf.toString();
	}

	private synchronized static void printStats() {
		sLog.debug("Thread count=" + getThreadCount() +
				"\t\t Max Thread count=" + maxThreadCount +
				"\t\t Average response time=" + getAverageResponseTime() +
				"\t\t for " + numOfRequests + " requests." +
				"\t\t Max response time(200 only)=" + maxResponseTime + "ms." +
				"\t\t Min response time(200 only)=" + minResponseTime + "ms." +
				getErrorCodes());

	}

	public synchronized static long getAverageResponseTime() {
		return responsetime / (numOfRequests == 0 ? 1 : numOfRequests);
	}

	public static synchronized void inc() {
		threadCounter++;
		maxThreadCount = Math.max(maxThreadCount, threadCounter);
		printStats();
	}

	public static synchronized void dec() {
		threadCounter--;
		deadThreadCount++;
		printStats();
	}

	public static synchronized void incError(int errorCode) {
		errorCounter[errorCode] = errorCounter[errorCode] + 1;
		if (errorCode != 200) {
			printStats();
		}
	}

	public static synchronized int getThreadCount() {
		return threadCounter;
	}


	public static String getPropertyString(java.util.Properties p) {
		StringBuffer buf = new StringBuffer();
		java.util.Enumeration e = p.propertyNames();
		while (e.hasMoreElements()) {
			String name = (String) e.nextElement();
			String value = p.getProperty(name);
			buf.append("\n\t").append(name).append("=").append(value);
		}
		return buf.toString();
	}

	public static String printRespHeader(Header[] headers) {
		StringBuffer buf = new StringBuffer("HEADERS:");
		for (int i = 0; i < headers.length; i++) {
			buf.append("\n\t");
			buf.append(headers[i].toExternalForm());
		}
		return buf.toString();
	}

	public static void main(String[] args) throws Exception {
		sLog.debug("Usage org.apache.catalina.cluster.test.client.JakartaClient\n\t" +
				"numberofthreads host port reps delay\n\t" + "Example:\n\t" +
				"java org.apache.catalina.cluster.test.client.JakartaClient 3 localhost 8080 1000 100");

		int numOfThreads = Integer.parseInt(args[0]);
		sLog.debug("numOfThreads : " + numOfThreads);
		String host = args[1];
		sLog.debug("Host : " + host);
		int port = Integer.parseInt(args[2]);
		sLog.debug("port : " + port);
		int reps = Integer.parseInt(args[3]);
		sLog.debug("reps : " + reps);
		long delay = Long.parseLong(args[4]);
		sLog.debug("delay : " + delay);
		String attrType = args[5];
		sLog.debug("attrType : " + attrType);
		sLog.debug("Cluster test started at " + new java.util.Date());
		for (int i = 0; i < numOfThreads; i++) {
			SessionReplicationClient t = new SessionReplicationClient("JakartaClient #" + i, host,
			port, reps, delay, attrType);
			t.setDaemon(false);

			t.start();
			Thread.currentThread().sleep(delay / 3);
		}
	}

}
