package com.clusterapp.listeners;

import java.util.Enumeration;
//import javax.servlet.ServletContext;
import javax.servlet.http.HttpSession;
import javax.servlet.http.HttpSessionAttributeListener;
import javax.servlet.http.HttpSessionBindingEvent;
import javax.servlet.http.HttpSessionEvent;
import javax.servlet.http.HttpSessionListener;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;

public final class ClusterAppSessionListener implements HttpSessionAttributeListener, 
		HttpSessionListener {
    private static Log sLog = LogFactory.getLog(ClusterAppSessionListener.class);
    private static int sessionCount = 0;

	public void sessionCreated(HttpSessionEvent event) {
		sLog.debug("HttpSession object is created @ " + new java.util.Date());
		sessionCount++;
		sLog.debug("sessionCreated: HttpSession count = " + sessionCount);
	}

	public void sessionDestroyed(HttpSessionEvent event) {
		sLog.debug("HttpSession object is removed @ " + new java.util.Date());
		if (sessionCount > 0) sessionCount--;
		sLog.debug("sessionDestroyed: HttpSession count = " + sessionCount);
	}

	public void attributeAdded(HttpSessionBindingEvent event) {
		sLog.debug("An attribute is added in the HttpSession object @ "
				+ new java.util.Date());
		printSessionInfo(event.getSession());
	}

	public void attributeRemoved(HttpSessionBindingEvent event) {
		sLog.debug("An attribute is removed in the HttpSession object @ "
				+ new java.util.Date());
		printSessionInfo(event.getSession());
	}

	public void attributeReplaced(HttpSessionBindingEvent event) {
		sLog.debug("An attribute is replaced in the HttpSession object @ "
				+ new java.util.Date());
		printSessionInfo(event.getSession());
	}

	// Utility Methods
	public static int getSessionCount() {
		return sessionCount;
	}
	public void printSessionInfo(HttpSession session) {
		sLog.debug("Session Id : " + session.getId());
		sLog.debug("Creation Time : " + session.getCreationTime());
		sLog.debug("Last Accessed Time : " + session.getLastAccessedTime());
		sLog.debug("New Session : " + session.isNew());
		sLog.debug("Session Attributes:");

		Enumeration attrNames = session.getAttributeNames();
		while (attrNames.hasMoreElements()) {
			String attrName = attrNames.nextElement().toString();
			String attrValue = session.getAttribute(attrName).toString();
			sLog.debug(attrName + "=" + attrValue);
		}
	}
}
// End of ClusterAppSessionListener.java
