package com.clusterapp.servlet;

import java.io.*;
import java.util.*;

import javax.servlet.*;
import javax.servlet.http.*;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;

import com.clusterapp.session.SessionAttributeFactory;
//import org.apache.catalina.cluster.test.ReplicationEvaluator;

public class ReplicationServlet extends HttpServlet {

	private static Log sLog = LogFactory.getLog(ReplicationServlet.class);

	private static final String CONTENT_TYPE = "text/plain";

//    String EVALUATOR_CLASS = "org.apache.catalina.cluster.test.server.SimpleReplicationEvaluator";

	//Initialize global variables
	public void init() throws ServletException {
		sLog.debug(getInitParameter("clusterNode") + "- init() method called.");
	}

	//Clean up resources
	public void destroy() {
		sLog.debug(getInitParameter("clusterNode") + "- destroy() method called.");
	}

	public void service(HttpServletRequest request, HttpServletResponse response) throws ServletException {
		sLog.debug(getInitParameter("clusterNode") + "- service() method called.");
		String strResponse = "";

		if ( request.getSession().isNew() && ( request.getParameter("sessionId") == null )) {
			sLog.debug("Evaluating new session");
			//getEvaluator().generateNewSessionRequest(request,response,true);
			strResponse = generateNewSessionRequest(request,response,true);
		} else {
			sLog.debug("Evaluating an existing session");
			//getEvaluator().evaluate(request,response);
			strResponse = evaluate(request,response);
		}
		sLog.debug("strResponse : " + strResponse);
		try {
			response.getOutputStream().write(strResponse.getBytes());
		} catch (IOException ioe) {
			sLog.error(ioe);
			throw new ServletException(ioe);
		}
	}

/*
	public ReplicationEvaluator getEvaluator() throws ServletException {
        try {
            return (ReplicationEvaluator) Thread.currentThread().
                getContextClassLoader().loadClass(EVALUATOR_CLASS).newInstance();
        } catch ( Exception x ) {
            throw new ServletException(x);
        }
    }
*/

	public synchronized String generateNewSessionRequest(HttpServletRequest request, HttpServletResponse response, boolean integritycheck) throws ServletException {
		if (integritycheck) {
			String id = request.getParameter("sessionId");
			if (id != null) {
				generateErrorReport(request, response,
						"Started a new session with an existing client session. Client session=" +
						id + " server session=" +
						request.getSession().getId());
				return "";
			}
		} else {
			request.getSession().invalidate();
			request.getSession();
		}
		response.setContentType(CONTENT_TYPE);
		request.getSession().setAttribute("counter","1");
		printHeaderInfo(request,response);
		return generateContinuePage(request,response);
	}

	public synchronized String evaluate(HttpServletRequest request, HttpServletResponse response) throws ServletException {
		response.setContentType(CONTENT_TYPE);
		/**
		 * This servlet does the following.
		 * 0. If the clients sends a reset request, then just send down the current state
		 * 1. It gets the session or creates one
		 * 2. Compare the session Id - always
		 * 3. If the session Id doesn't match, then generate error page
		 * 4. Compare all the attributes that we are expecting, if fails
		 *    generate error page
		 * 5. If it works, update the counter
		 * 6. Execute the SessionCommand
		 * 7. Generate the response page
		 *
		 * Parameter Names expected:
		 * sessionId
		 * param_xx - where xx is a counter on the session attribute name
		 * value_xx - where xx is a counter on the session attribute value
		 * counter  - the number of requests
		 * sessionCommand - can have one of the values "add", "remove", "nothing"
		 * sessionCommandName - the name of the attribute to be added/removed, has to be param_xx
		 * sessionCommandValue - the value of the attribute to be added, has to be value_xx
		 *
		 */
		//check if we want to reset
		if ( "reset".equals(request.getParameter("sessionCommand"))) {
			return generateNewSessionRequest(request,response,false);
		}

		String id = request.getParameter("sessionId");
		sLog.debug("id = " + id);
		HttpSession session = request.getSession();
		sLog.debug("session.getId() = " + session.getId());
		if (!session.getId().equals(id)) {
			generateErrorReport(request,response,"Invalid session id server=" + session.getId() + " client=" + id);
			return "";
		}
		id = session.getId();

		String counter = request.getParameter("counter");
		sLog.debug("counter = " + counter);
		String c = (String)session.getAttribute("counter");
		sLog.debug("c = " + c);
		if (c == null || (!c.equals(counter))) {
			generateErrorReport(request,response,"Invalid counter id server=" + c + " client=" + counter);
			return "";
		} //end if
		c = String.valueOf(Integer.parseInt(c)+1);
		session.setAttribute("counter",c);

		//loop through client parameters
		java.util.Enumeration e = request.getParameterNames();
		while (e.hasMoreElements()) {
			String name = (String) e.nextElement();
			if (name.startsWith("param_")) {
				String value = request.getParameter(name);
				Object svalue = session.getAttribute(name);
				if (svalue == null || value == null) {
					generateErrorReport(request,response,"Session data mismatch[1] client name="+name+" value="+value+" server value="+svalue);
					return "";
				}
				if (!value.equals(svalue.toString())) {
					generateErrorReport(request,response,"Session data mismatch[2] client name="+name+" value="+value+" server value="+svalue);
					return "";
				}
			}//end if
		}//while

		//loop through server attributes to match
		e = session.getAttributeNames();
		while (e.hasMoreElements()) {
			String name = (String)e.nextElement();
			if (name.startsWith("param_")) {
				String value = request.getParameter(name);
				Object svalue = session.getAttribute(name);
				if (svalue == null || value == null) {
					generateErrorReport(request,response,"Session data mismatch[3] client name="+name+" value="+value+" server value="+svalue);
					return "";
				}
				if (!value.equals(svalue.toString())) {
					generateErrorReport(request,response,"Session data mismatch[4] client name="+name+" value="+value+" server value="+svalue);
					return "";
				}
			}//end if
		}//while

		String command = request.getParameter("sessionCommand");
		sLog.debug("**** command: " + command);

		String attrName = "";
		String attrValue = "";
		if (command != null) {
			if ("add".equals(command)) {
				//session.setAttribute(request.getParameter("sessionCommandName"),
				//		SessionAttributeFactory.getSessionAttribute(request.getParameter("sessionCommandValue")));
				//System.out.println("["+id+"] Added attribute="+session.getAttribute(request.getParameter("sessionCommandName")));

				// get the attributes to store in session
				Enumeration paramNames = request.getParameterNames();
				while (paramNames.hasMoreElements()) {
					String paramName = paramNames.nextElement().toString();
					//if (paramName.indexOf("sessionAttrName") != -1) {
					if (paramName.indexOf("sessionAttrValue") != -1) {
						sLog.debug("paramName: " + paramName);
						attrName = paramName;
						attrValue = request.getParameter(paramName);
						sLog.debug("attrValue: " + attrValue);
						//session.setAttribute(attrName, SessionAttributeFactory.getSessionAttribute(attrValue));
						session.setAttribute(attrName, SessionAttributeFactory.getSessionAttribute(attrValue));
						sLog.debug("[" + id + "] Added attribute: Name= " + attrName +
								" Value=" + session.getAttribute(attrName));
					}
				}

			} else if ("remove".equals(command)) {
				if ( session.getAttribute(request.getParameter("sessionCommandName")) == null ) {
					generateErrorReport(request,response,"Unable to remove attribute, nonexistent client name="+request.getParameter("sessionCommandName"));
					return "";
				} else {
					//                    System.out.println("["+id+"] Attribute removed name=" +
					//                                       request.getParameter("sessionCommandName")+" Value="+
					//                                       session.
					//                                       getAttribute(request.getParameter("sessionCommandName")));
					session.removeAttribute(request.getParameter("sessionCommandName"));
				}//end if
			}//end if
		}//end if
		printHeaderInfo(request,response);
		return generateContinuePage(request,response);
	}

	public void generateErrorReport(HttpServletRequest request, HttpServletResponse response, String message) throws ServletException {

		sLog.debug(getInitParameter("clusterNode") + "- generateErrorReport() method called.");

		// When an error happens print out headers and more
		StringBuffer strError = new StringBuffer("Error request on session:").append(message);
		strError.append("\nHEADERS:");
		java.util.Enumeration e = request.getHeaderNames();
		while ( e.hasMoreElements() ) {
			String name = (String)e.nextElement();
			strError.append("\n\t").append(name).append("=").append(request.getHeader(name));
		}
		strError.append("\nPARAMETERS:");
		e = request.getParameterNames();
		while ( e.hasMoreElements() ) {
			String name = (String)e.nextElement();
			strError.append("\n\t").append(name).append("=").append(request.getParameter(name));
		}
		strError.append("\n");
		sLog.debug(strError);
		try {
			String msg = "["+java.net.InetAddress.getLocalHost()+"]" +message;
			response.sendError(501, msg);
		} catch ( Exception x ) {
			x.printStackTrace();
		}
	}

	public String generateContinuePage(HttpServletRequest request, HttpServletResponse response) throws ServletException {
		/**
		 * generates a java.util.Properties format response. with the following parameters
	    * sessionId
		 * param_xx - where xx is a counter on the session attribute name
		 * value_xx - where xx is a counter on the session attribute value
		 * counter  - the number of requests
		 */

		sLog.debug(getInitParameter("clusterNode") + "- generateContinuePage() method called.");

		java.util.Properties p = new java.util.Properties();
		java.util.Enumeration e = request.getSession().getAttributeNames();
		java.io.ByteArrayOutputStream out = new java.io.ByteArrayOutputStream();
		p.setProperty("sessionId",request.getSession().getId());
		while (e.hasMoreElements()) {
			String name = (String)e.nextElement();
			p.setProperty(name,request.getSession().getAttribute(name).toString());
		}
		try {
			p.store(out, "");
			return new String(out.toByteArray());
		} catch (Exception x) {
			throw new ServletException(x);
		}
	}

	public void printHeaderInfo(HttpServletRequest request, HttpServletResponse response) throws ServletException {

		sLog.debug(getInitParameter("clusterNode") + "- printHeaderInfo() method called.");

		// When an error happens print out headers and more
		StringBuffer strHeader = new StringBuffer("Request Header Info:");
		strHeader.append("\nHEADERS:");
		java.util.Enumeration e = request.getHeaderNames();
		while ( e.hasMoreElements() ) {
			String name = (String)e.nextElement();
			strHeader.append("\n\t").append(name).append("=").append(request.getHeader(name));
		}
		strHeader.append("\nPARAMETERS:");
		e = request.getParameterNames();
		while ( e.hasMoreElements() ) {
			String name = (String)e.nextElement();
			strHeader.append("\n\t").append(name).append("=").append(request.getParameter(name));
		}
		strHeader.append("\n");
		sLog.debug(strHeader.toString());
	}

}
